#ifndef lint
static char Rcs_Id[] =
	"$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/debugio.c,v 2.17 2000/04/13 03:19:35 johnh Exp $";
#endif

/*
 * $Log: debugio.c,v $
 * Revision 2.17  2000/04/13 03:19:35  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.16  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.15  1998/07/30 20:42:24  geoff
 * Only compile dumpqueues if DUMP is defined.
 *
 * Revision 2.14  1995/11/27  01:59:36  geoff
 * Properly declare dumpproc as a static.
 *
 * Revision 2.13  1995/05/11  05:05:19  geoff
 * Minor RCS header changes.
 *
 * Revision 2.12  86/11/29  19:50:43  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.11  86/03/31  22:57:49  geoff
 * Add code to dump new fields used in supporting queue limits.
 * 
 * Revision 2.10  86/03/30  16:52:24  geoff
 * Add code to dump incremented ranges intelligently
 * 
 * Revision 2.9  86/03/30  14:33:20  geoff
 * Add dumpqueues.
 * 
 * Revision 2.8  86/03/28  14:41:33  geoff
 * Change DEBUG to the DUMP option so that they are separable.  Add code to
 * dump the new fields.
 * 
 * Revision 2.7  86/03/25  19:56:05  geoff
 * Add support for FF_ATJOB and EF_ATJOB, and handle null ce_command fields.
 * 
 * Revision 2.6  86/01/04  16:16:34  geoff
 * Clean up some minor formatting problems.
 * 
 * Revision 2.5  85/11/25  16:51:51  geoff
 * Fix a lot of lint complaints.
 * 
 * Revision 2.4  85/11/24  16:12:16  geoff
 * Add dumpproc and dumpflags.  Expand flag dumping for all tables.
 * Cast a bunch of putc's and fprintf's to void to stop lint complaints.
 * 
 * Revision 2.3  85/11/05  00:35:55  geoff
 * Make changes to reflect the new flag namings.
 * 
 * Revision 2.2  85/11/02  19:07:09  geoff
 * Modify dumpent to always put out flags.  Modify ascuid to make use of the
 * user table to deduce the user name when possible.
 * 
 * Revision 2.1  85/10/25  02:04:22  geoff
 * Add support for new format crontabs and symbolic names
 * 
 * Revision 1.2  85/10/13  21:47:22  geoff
 * Fix some lint complaints
 * 
 * Revision 1.1  85/10/12  01:33:32  geoff
 * Break it up into separate modules
 * 
 */

/*
 * Debug I/O routines for the extended clock daemon.
 */

#include "cron.h"
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <time.h>

#ifdef DUMP
void		dumpall ();		/* Dump all data structures */
void		dumpfile ();		/* Dump a file descriptor */
void		dumpuser ();		/* Dump a user table entry */
void		dumpent ();		/* Dump a command entry */
static void	dumpproc ();		/* Dump a process table entry */
static void	dumpqueues ();		/* Dump pending-process queues */
static void	dumpflags ();		/* Dump a flags word */
static void	dumpafield ();		/* Dump a symbolic time field */
static void	dumpnfield ();		/* Dump a numeric time field */
#endif
char *		ascgid ();		/* Return ascii form of a GID */
char *		ascuid ();		/* Return printable form of UID */

extern char *	daynames[];		/* Table of names of the days */
extern struct cronent enthead;		/* Head of cronent chain */
extern struct cronfile filehead;	/* Head of cronfile chain */
extern struct group * getgrgid ();	/* Get a group entry by group id */
extern struct passwd * getpwuid ();	/* Get a password entry by uid */
extern char *	monames[];		/* Tables of names of the months */
extern struct childproc prochead;	/* Head of childproc chain */
extern time_t	time ();		/* Return system time */
extern char *	timenonl ();		/* Return ASCII time w/o newline */
extern struct cronuser userhead;	/* Head of cronuser chain */
extern int	yearbase;		/* Base year for bit masks */

#ifdef DUMP
void dumpall ()				/* Dump all the neat structures */
    {
    reg1 struct cronfile *	filep;	/* Current file table entry */
    reg2 struct cronent *	entp;	/* Current cron table entry */
    reg3 struct childproc *	procp;	/* Current child process entry */
    reg4 struct cronuser *	userp;	/* Current user table entry */

    (void) fseek (stderr, 0L, 2);
    if (userhead.cu_next == &userhead)
	(void) fprintf (stderr, "user table is empty!\n");
    else
	{
	(void) fprintf (stderr, "\nUser table as of %s:\n",
	  timenonl (time ((long *) NULL)));
	(void) fprintf (stderr,
	  "flags\tpr pl ql ni\tuid\tgid\tname\t\thome dir\n");
	for (userp = userhead.cu_next;
	  userp != &userhead;
	  userp = userp->cu_next)
	    dumpuser (userp);
	}
    if (filehead.cf_next == &filehead)
	(void) fprintf (stderr, "file table is empty!\n");
    else
	{
	(void) fprintf (stderr, "\nFile table as of %s:\n",
	  timenonl (time ((long *) NULL)));
	(void) fprintf (stderr, "flags\tuid\trtime\t\t\t\tname\n");
	for (filep = filehead.cf_next;
	  filep != &filehead;
	  filep = filep->cf_next)
	    dumpfile (filep);
	}
    if (enthead.ce_next == &enthead)
	(void) fprintf (stderr, "\nentry table is empty\n");
    else
	{
	(void) fprintf (stderr, "\nEntry table:\n");
	(void) fprintf (stderr, "file\tuid\tgid\tflags\ttime\t\tnext-time\n");
	for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	    dumpent (entp);
#ifdef LOC_CRON
	(void) fprintf (stderr, "\nEntry table flags:\n\tw: EF_WAIT (run during startup and wait)\n\t&: EF_SPAWN (run during startup and don't wait)\n\ts: EF_SEQSPAWN (Run in seq during bckgrnd startup)\n\ta: EF_RUNALL (Run all catchups of this cmd)\n\tm: EF_ONLYCATCHUP (Run only during catchup mode)\n\td: EF_MARKDEL (Entry is marked for final delete)\n\tD: EF_DELETED (Entry is logically deleted)\n\tA: EF_ATJOB  (Entry is an at job, not cron)\n\th: EF_HASRUN (Job has run during catchup)\n\t@: EF_IS_AT_LOCATION (job is at suitable location to run)\n\tW: EF_WAS_AT_LOCATION ( " " ...before last move)\n\tM: EF_ONLY_ON_MOVE (should this job run only on moves?)\n\tF: EF_ONLY_ON_MOVE_FIRED (are at jobs already running?)\n");
#endif /* LOC_CRON */
	}
    (void) fprintf (stderr, "\n");
    if (prochead.cp_next == &prochead)
	(void) fprintf (stderr, "process table is empty\n");
    else
	{
	(void) fprintf (stderr, "\nProcess table:\n");
	(void) fprintf (stderr, "pid\topfx\tepfx\tfbase\t\t\tcommand\n");
	for (procp = prochead.cp_next;
	  procp != &prochead;
	  procp = procp->cp_next)
	    dumpproc (procp);
	}
    dumpqueues ();			/* Dump pending-process queues */
    (void) fprintf (stderr, "\n");
    }

void dumpuser (userp)			/* Dump a user table entry */
    reg1 struct cronuser *	userp;	/* Current user table entry */
    {
    dumpflags (userp->cu_flags, "dnJ");
    (void) fprintf (stderr, "\t%2d %2d %2d %2d\t%s\t%s\t%s\t%s\n",
      userp->cu_procs, userp->cu_proclim, userp->cu_qlim, userp->cu_nice,
      ascuid (userp->cu_uid),
      ascgid (userp->cu_gid), userp->cu_name, userp->cu_home);
    }

void dumpfile (filep)			/* Dump a file descriptor */
    reg1 struct cronfile *	filep;	/* Current file table entry */
    {
    dumpflags (filep->cf_flags, "mdA");
    (void) fprintf (stderr, "\t%s\t%s\t%s\n", ascuid (filep->cf_user->cu_uid),
      timenonl (filep->cf_rtime), filep->cf_name);
    }

void dumpent (entp)			/* Dump a command entry */
    reg1 struct cronent *	entp;	/* Current cron table entry */
    {

    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "%s\t%s\t%s\t", entp->ce_file->cf_name,
      ascuid (entp->ce_uid), ascgid (entp->ce_gid));
    if (entp->ce_flags == 0)
	(void) putc ('-', stderr);
    else {
#ifdef LOC_CRON
	dumpflags (entp->ce_flags, "w&samdDAh@WMF");
#else /* ! LOC_CRON */
	/* next line had a bug:  missing D entry */
	dumpflags (entp->ce_flags, "w&samdDAh");
#endif /* LOC_CRON */
    };
    (void) putc ('\t', stderr);
    dumpafield (entp->ce_wkday, daynames, 7);
    dumpafield (entp->ce_month, monames, 12);
    dumpnfield (entp->ce_moday, 31, 1, 0, ' ');
    dumpnfield (entp->ce_year, YEARMAX, yearbase, 0, ' ');
    dumpnfield (entp->ce_hour, 24, 0, 1, ':');
    dumpnfield (entp->ce_min, 60, 0, 1, ' ');
#ifdef LOC_CRON
    if (entp->ce_location)
	    (void) fprintf(stderr, "@%s ",  entp->ce_location);
#endif /* LOC_CRON */
    (void) fprintf (stderr, "%24.24s ", ctime (&entp->ce_rtime));
    (void) fprintf (stderr, "%s\n", entp->ce_command ? entp->ce_command : "");
    }

static void dumpproc (procp)			/* Dump a process entry */
    reg1 struct childproc *	procp;		/* Entry to dump */
    {
    (void) fprintf (stderr, "%d\t%d\t%d\t%s\t%s\n",
      procp->cp_pid, procp->cp_oprefix, procp->cp_eprefix, procp->cp_fbase,
      procp->cp_ent->ce_command ? procp->cp_ent->ce_command : "(at job)");
    }

static void dumpflags (flags, chars)		/* Dump a flags word */
    reg2 int		flags;			/* Word to dump */
    reg1 char *		chars;			/* String of flag chars */
    {

    while (*chars)
	{
#ifdef LOC_CRON
	if ((flags & 1) && *chars != ' ')
	    (void) putc (*chars, stderr);
#else /* ! LOC_CRON */
	if (flags & 1)
	    (void) putc (*chars, stderr);
#endif /* LOC_CRON */
	flags >>= 1;
	chars++;
	}
    }

static void dumpafield (field, table, size)	/* Dump a symbolic field */
    reg1 MASKTYPE *	field;			/* Field to dump */
    reg5 char *		table[];		/* Table of value names */
    reg6 int		size;			/* Size of the field */
    {
    reg2 int		curbit;			/* Current bit no. in field */
    reg4 int		incr;			/* Increment for / option */
    reg3 int		start;			/* Start bit of nonzero run */
    int			needcomma;		/* NZ if commas are needed */

    start = -1;
    needcomma = 0;
    for (curbit = 0;  curbit < size;  curbit++)
	{
	if (field[MASKBYTE (curbit)] & MASKBIT (curbit))
	    {					/* Nonzero bit */
	    if (start == -1)			/* Begin a run of ones? */
		start = curbit;			/* If so, record start */
	    }
	else					/* A zero bit */
	    {
	    if (start != -1)			/* Were we in a run? */
		{
		if (needcomma)
		    (void) putc (',', stderr);
		if (curbit - start == 1)
		    {
		    /*
		     * A 1-bit run.  Check if there is a repeated pattern
		     * that is worth dumping.
		     */
		    for (incr = 0, curbit++;  curbit < size;  curbit++)
			{
			if (field[MASKBYTE (curbit)] & MASKBIT (curbit))
			    {
			    /*
			     * We have found another one bit after a
			     * string of zeroes.  See if it's a
			     * repeated pattern.
			     */
			    if (incr == 0)
				incr = curbit - start;
			    else if ((curbit - start) % incr != 0)
				break;
			    }
			else
			    {
			    if (incr  &&  (curbit - start) % incr == 0)
				{
				/* It's supposed to be 1, but it's not */
				curbit--;
				break;
				}
			    }
			}
		    if (curbit == size)
			curbit--;
		    if (incr)
			curbit -= (curbit - start) % incr;
		    if (incr == 0  ||  (curbit - start) < 2 * incr)
			{
			/* Not worth representing with an increment */
			(void) fprintf (stderr, "%s", table[start]);
			curbit = start + 1;	/* Reset for more looping */
			}
		    else
			(void) fprintf (stderr, "%s-%s/%d", table[start],
			  table[curbit], incr);
		    }
		else if (curbit - start == 2)
		    {
		    curbit -= 2;		/* Re-scan this next pass */
		    (void) fprintf (stderr, "%s", table[start]);
		    }
		else
		    (void) fprintf (stderr, "%s-%s", table[start],
		      table[curbit - 1]);
		needcomma = 1;
		start = -1;
		}
	    }
	}
    if (start != -1)
	{
	if (needcomma)
	    (void) putc (',', stderr);
	if (start == 0  &&  curbit == size)	/* All vals == null field */
	    return;
	else if (curbit - start == 1)
	    (void) fprintf (stderr, "%s", table[start]);
	else if (curbit - start == 2)
	    (void) fprintf (stderr, "%s,%s", table[start], table[start + 1]);
	else
	    (void) fprintf (stderr, "%s-%s", table[start], table[curbit - 1]);
	}
    (void) putc (' ', stderr);
    }

    /* Dump a numeric field */
static void dumpnfield (field, size, base, star, pad)
    reg1 MASKTYPE *	field;			/* Field to dump */
    reg6 int		size;			/* Size of the field */
    reg5 int		base;			/* Base of field's indexing */
    int			star;			/* NZ to get star, not null */
    char		pad;			/* Char to follow field */
    {
    reg2 int		curbit;			/* Current bit no. in field */
    reg3 int		incr;			/* Increment for / option */
    reg4 int		start;			/* Start bit of nonzero run */
    int			needcomma;		/* NZ if commas are needed */

    start = -1;
    needcomma = 0;
    for (curbit = 0;  curbit < size;  curbit++)
	{
	if (field[MASKBYTE (curbit)] & MASKBIT (curbit))
	    {					/* Nonzero bit */
	    if (start == -1)			/* Begin a run of ones? */
		start = curbit;			/* If so, record start */
	    }
	else					/* A zero bit */
	    {
	    if (start != -1)			/* Were we in a run? */
		{
		if (needcomma)
		    (void) putc (',', stderr);
		if (curbit - start == 1)
		    {
		    /*
		     * A 1-bit run.  Check if there is a repeated pattern
		     * that is worth dumping.
		     */
		    for (incr = 0, curbit++;  curbit < size;  curbit++)
			{
			if (field[MASKBYTE (curbit)] & MASKBIT (curbit))
			    {
			    /*
			     * We have found another one bit after a
			     * string of zeroes.  See if it's a
			     * repeated pattern.
			     */
			    if (incr == 0)
				incr = curbit - start;
			    else if ((curbit - start) % incr != 0)
				break;
			    }
			else
			    {
			    if (incr  &&  (curbit - start) % incr == 0)
				{
				/* It's supposed to be 1, but it's not */
				curbit--;
				break;
				}
			    }
			}
		    if (curbit == size)
			curbit--;
		    if (incr)
			curbit -= (curbit - start) % incr;
		    if (incr == 0  ||  (curbit - start) < 2 * incr)
			{
			/* Not worth representing with an increment */
			(void) fprintf (stderr, "%d", start + base);
			curbit = start + 1;	/* Reset for more looping */
			}
		    else
			(void) fprintf (stderr, "%d-%d/%d", start + base,
			  curbit + base, incr);
		    }
		else if (curbit - start == 2)
		    {
		    curbit -= 2;		/* Re-scan this next pass */
		    (void) fprintf (stderr, "%d", start + base);
		    }
		else
		    (void) fprintf (stderr, "%d-%d",
		      start + base, curbit + base - 1);
		needcomma = 1;
		start = -1;
		}
	    }
	}
    if (start != -1)
	{
	if (needcomma)
	    (void) putc (',', stderr);
	if (start == 0  &&  curbit == size)	/* Special-case '*' */
	    {
	    if (star)
		(void) putc ('*', stderr);
	    else
		return;				/* Completely null */
	    }
	else if (curbit - start == 1)
	    (void) fprintf (stderr, "%d", start + base);
	else if (curbit - start == 2)
	    (void) fprintf (stderr, "%d,%d", start + base, start + base + 1);
	else
	    (void) fprintf (stderr, "%d-%d", start + base, curbit + base - 1);
	}
    (void) putc (pad, stderr);
    }
#endif

char * ascgid (gid)			/* Return printable form of GID */
    int			gid;		/* GID to be converted */
    {
    static char		buf[10];	/* Space for converted GID */
    struct group *	grent;		/* Password entry for GID */

    grent = getgrgid (gid);
    if (grent == NULL)
	{
	(void) sprintf (buf, "%d", gid);
	return buf;
	}
    else
	return grent->gr_name;		/* Return user name */
    }

char * ascuid (uid)			/* Return printable form of UID */
    int			uid;		/* UID to be converted */
    {
    static char		buf[10];	/* Space for converted UID */
    struct passwd *	pwent;		/* Password entry for UID */
    reg1 struct cronuser *
			userp;		/* Current user table entry */

    for (userp = userhead.cu_next;
      userp != &userhead;
      userp = userp->cu_next)
	{
	if (userp->cu_uid == uid)
	    return userp->cu_name + EL_LOGNAME;
	}
    pwent = getpwuid (uid);
    if (pwent == NULL)
	{
	(void) sprintf (buf, "UID %d", uid);
	return buf;
	}
    else
	return pwent->pw_name;		/* Return user name */
    }
#ifdef DUMP

static void	dumpqueues ()		/* Dump pending-process queues */
    {
    reg3 int			needheader; /* NZ if head is still needed */
    reg1 struct cmdqueue *	queuep;	/* Current queue entry */
    reg2 struct cronuser *	userp;	/* Current user */

    for (userp = userhead.cu_next, needheader = 1;
      userp != &userhead;
      userp = userp->cu_next)
	{
	for (queuep = userp->cu_queue.cq_next;
	  queuep != &userp->cu_queue;
	  queuep = queuep->cq_next)
	    {
	    if (needheader)
		{
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr, "\nUser\tflags\tpending command\n");
		needheader = 0;
		}
	    (void) fprintf (stderr, "%s\t",
	      ascuid (queuep->cq_ent->ce_file->cf_user->cu_uid));
	    dumpflags (queuep->cq_flags, "l");
	    (void) fprintf (stderr, "\t%s\n",
	      queuep->cq_ent->ce_command);
	    }
	}
    if (needheader)
	(void) fprintf (stderr, "\nNo pending commands\n");
    }
#endif /* DUMP */
