#ifndef lint
static char Rcs_Id[] =
	"$Header: /h/guests/fmg/johnh/GEOFF_CVS/xcron/parse.c,v 2.28 2000/04/13 03:19:36 johnh Exp $";
#endif

/*
 * $Log: parse.c,v $
 * Revision 2.28  2000/04/13 03:19:36  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.27  2000/02/08 08:56:42  geoff
 *
 * Integrate various changes from Geoff's branch into the now-master CVS
 * repository.  As part of this integration, I cleaned up some, but not
 * all, of the inconsistent coding style introduced by the LOC_CRON
 * changes.  Important and substantive changes include:
 *
 *     Configure	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     Makefile.SH	Tiny cleanup to an expr command.
 *     config.SH	Add errlist_in_stdio and getitimer/sigprocmask support.
 *     errmsg.c	Add errlist_in_stdio support.
 *     mainloop.c	Use setjmp/longjmp to ensure that we are blown off the
 * 		select when we catch an alarm, even if we are running
 * 		on a system that has reliable signals.  Also add
 * 		reliable-signal support that properly resets the
 * 		signal mask in the longjmp case.
 *     parse.c	Fix a bug that could cause crontabs to be accepted
 * 		with the wrong UID.  Be a bit more careful about
 * 		parsing years.
 *     runcmd.c	Add setjmp/longjmp and reliable-signal support.  (See
 * 		mainloop.c).
 *     xcron.c	Use setitimer to make absolutely sure that all alarms
 * 		are disabled at startup.
 *
 * Revision 2.26  1999/07/27 01:22:00  johnh
 * location support for at jobs
 *
 * Revision 2.25  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.24  1998/12/17 23:10:01  geoff
 * Don't pass unrepresentable at-job times to nextrun.
 *
 * Revision 2.23  1998/07/30 21:42:50  geoff
 * Add support for entry-size limits.
 *
 * Revision 2.22  1998/07/30 20:42:27  geoff
 * Many changes, accumulated over far too much time: add partial support
 * for resource-exhaustion limits.  Support long names.
 *
 * Revision 2.21  1995/11/27  06:13:38  geoff
 * Fix a couple of error prints that crashed the daemon!
 *
 * Revision 2.20  1995/11/27  01:59:44  geoff
 * Save errno when putting out error messages.  Add support for old-style
 * crontab lines.
 *
 * Revision 2.19  1995/05/11  05:05:44  geoff
 * Minor RCS header changes.
 *
 * Revision 2.18  86/12/01  14:46:03  geoff
 * Fix several lint complaints.
 * 
 * Revision 2.17  86/11/29  19:52:47  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.16  86/11/16  00:58:50  geoff
 * Disallow the 'w' flag in user crontabs.
 * 
 * Revision 2.15  86/04/17  21:59:58  geoff
 * Modify some log messages to support the crontab -L switch, and add a log
 * message for queuing of at jobs.
 * 
 * Revision 2.14  86/04/02  00:02:18  geoff
 * Take out the third parameter to mailto.
 * 
 * Revision 2.13  86/04/01  19:03:00  geoff
 * Remove an obsolete line of code that cleared flag bits in non-startup mode.
 * 
 * Revision 2.12  86/03/31  22:58:27  geoff
 * Add support for queue limits.
 * 
 * Revision 2.11  86/03/30  15:41:30  geoff
 * Switch debugging output to the DBPRINTn macros when possible
 * 
 * Revision 2.10  86/03/30  14:33:46  geoff
 * Add code to sort the entry list by time as it is read in.
 * 
 * Revision 2.9  86/03/28  14:42:01  geoff
 * Add support for new fields.  Add code to calculate the next-run-time
 * of each entry read and to insert it into the list in sorted order.
 * Add the insent routine.  Reject crontab time increments less than 1.
 * Fix some lint complaints.
 * 
 * Revision 2.8  86/03/25  19:56:21  geoff
 * Add support for at jobs.
 * 
 * Revision 2.7  86/01/04  16:16:58  geoff
 * Use tempfile in place of mktemp, and initialize each user's command-queue
 * to empty.
 * 
 * Revision 2.6  85/12/21  15:15:32  geoff
 * Add #ifdef's for ABSDIRS
 * 
 * Revision 2.5  85/11/25  16:52:14  geoff
 * Fix a bunch of lint complaints.  Fix the handling of errorok to be correct.
 * 
 * Revision 2.4  85/11/24  16:13:28  geoff
 * Fix some lint complaints.  Remove the obsolete 'uid' argument, and
 * associated code, from addfile.  Remove support for the EF_USERTAB and
 * EF_NEWFILES flags.
 * 
 * Revision 2.3  85/11/05  00:36:20  geoff
 * Mail syntax error messages to the owner of the cron table, instead of
 * putting them into the cron log.  Add support for separate garbage
 * collection.
 * 
 * Revision 2.2  85/11/03  00:03:13  geoff
 * Make the chdir's for the various directories permanent.  Modify the setting
 * of the cu_name and cu_home variables to make them usable for environments.
 * 
 * Revision 2.1  85/10/25  02:03:38  geoff
 * Add support for new format crontabs and symbolic names
 * 
 * Revision 1.1  85/10/12  01:33:35  geoff
 * Break it up into separate modules
 * 
 */

#include "cron.h"
#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <time.h>

struct cronfile * addfile ();		/* Add a file to the cronfile list */
struct cronuser * adduser ();		/* Add a user to the user list */
char *		strcasn ();		/* Copy & cat strs to alloc'ed area */
void		delfile ();		/* Delete entries from a given file */
void		readall ();		/* Read in all files */
void		readfile ();		/* Read a file into cronent lists */
static struct cronent * readent ();	/* Read one command descriptor */
static void	makeatent ();		/* Make an at-job entry */
static void	decodeatfile ();	/* Turn an at-job name into a time */
void		insent ();		/* Insert an entry into the list */
static char *	getnonblank ();		/* Get a nonblank string */
static char *	getcommand ();		/* Get a command string */
static char *	parseprep ();		/* Prepare to parse new field */
static void	getafield ();		/* Convert a symbolic time field */
static void	getnfield ();		/* Convert a numeric time field */
static void	skipblanks ();		/* Skip blanks but not linefeeds */
static void	makelower ();		/* Make a buffer lowercase */
static int	prefixnum ();		/* See if string has a known prefix */

/*
 * Local variables
 */
char *		curfile = NULL;		/* Current file's name */
int		curline = 0;		/* Line number in current file */
char *		daynames[7] = DAYNAMES;	/* Day names, in local language */
FILE *		errfile;		/* File to write error messages to */
int		errfull;		/* NZ if err file has msgs in it */
char		errname[] = TEMPBASE;	/* Error file name */
char *		monames[12] = MONTHNAMES; /* Month names, in local language */
int		yearbase;		/* Base of year calculations */

extern char *	ascgid ();		/* Return ASCII form of a GID */
extern char *	ascuid ();		/* Return ASCII form of a UID */
extern void	collect ();		/* Garbage collection routine */
extern char *	curfile;		/* Current file's name */
extern int	curline;		/* Line number in current file */
#ifdef DEBUG
extern short	debug;			/* Nonzero to get debugging output */
extern void	dumpall ();		/* Dump all data structures */
extern void	dumpfile ();		/* Dump a file descriptor */
extern void	dumpent ();		/* Dump a command entry */
#endif
extern struct cronent enthead;		/* Head of cronent chain */
extern char *	errmsg ();		/* Return an error message string */
extern struct cronfile filehead;	/* Head of cronfile chain */
extern void	free ();		/* Return memory to the allocator */
extern struct group * getgrnam ();	/* Get group entry by name */
extern struct passwd * getpwnam ();	/* Get password entry by name */
extern struct passwd * getpwuid ();	/* Get a password entry by uid */
extern void	mailto ();		/* Mail a file to a user */
extern char *	malloc ();		/* Memory allocator */
extern char *	memset ();		/* Set memory to a constant value */
extern time_t	nextrun ();		/* Figure when an entry next runs */
extern char *	printable ();		/* Return printable version of char */
extern short	re_collect;		/* NZ if collection is needed later */
extern char *	realloc ();		/* Reallocate memory */
extern char *	strcat ();		/* Concatenate strings */
extern char *	strchr ();		/* Find character in string */
extern char *	strcpy ();		/* Copy strings */
extern void	synerr ();		/* Issue syntax error message */
extern char *	tempfile ();		/* Make a temporary file name */
extern time_t	time ();		/* Return system time */
extern char *	timenonl ();		/* Return a time w/o newline */
extern struct cronuser userdefaults;	/* Default limit values for a user */
extern struct cronuser userhead;	/* Head of cronuser chain */

	/* Add a file to the cronfile list */
struct cronfile * addfile (name, flags)
    reg3 char *			name;	/* Pathname of the file to add */
    reg4 int			flags;	/* Flags that apply to the file */
    {
    reg1 struct cronfile *	cfile;	/* Cronfile entry we're examining */
    int				day;	/* Day at job would run */
    reg3 int			filecount; /* Number of at files user has */
    int				hour;	/* Hour at job would run */
#ifdef LOC_CRON
    char *			loc;    /* location for job (only for ATLOC) */
#endif /* LOC_CRON */
    int				minute;	/* Minute at job would run */
    int				month;	/* Month at job would run */
    reg2 struct cronfile *	newfile; /* Cronfile entry we're creating */
    struct passwd *		pwent;	/* User's password file entry */
    int				saveerr; /* Place to save errno b4 clobber */
    struct stat			statbuf; /* Unix status of the file */
    reg4 struct cronuser *	user;	/* User-table entry for file owner */
    int				year;	/* Year at job would run */

    DBPRINT2 (1, "addfile (%s, %x)\n", name, flags);
    /* Prevent duplications */
    for (cfile = filehead.cf_next;
      cfile != &filehead;
      cfile = cfile->cf_next)
	{
	if (strcmp (name, cfile->cf_name) == 0)
	    {
	    DBPRINT0 (10, "addfile:  already there, clearing MARKDEL\n");
	    cfile->cf_flags &= ~(FF_MARKDEL | FF_DELETED);
					/* Make sure it isn't del'ed */
	    return cfile;
	    }
	}
    /*
     * Mode must have SUID and IREAD, but not SGID, SVTX, any EXEC, or
     * world write permissions.
     */
    if (flags & FF_ATJOB)
	(void) chdir (ATJOBDIR);
    else
	(void) chdir (CRONTABDIR);
    if (stat (name, &statbuf) < 0)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "addfile:  couldn't stat %s:  %s\n", name,
	  errmsg (saveerr));
	(void) chdir (LIBDIR);
	return NULL;			/* Don't insert the file */
	}
    (void) chdir (LIBDIR);
    if ((statbuf.st_mode & MUSTHAVE) != MUSTHAVE
      ||  (statbuf.st_mode & CANTHAVE) != 0)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "addfile:  %s has illegal mode 0%o\n",
	  name, statbuf.st_mode);
	return NULL;
	}
    if (flags & FF_ATJOB)
	{
	if ((user = adduser ((int) statbuf.st_uid)) == NULL)
	    return NULL;
	}
    else
	{
	pwent = getpwnam (name);
	if (pwent == NULL)
	    {
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "adduser:  couldn't find password entry for user %s\n", name);
	    return NULL;
	    }
	if ((user = adduser ((int) pwent->pw_uid)) == NULL)
	    return NULL;
	}
    for (filecount = 0, cfile = filehead.cf_next;
      cfile != &filehead;
      cfile = cfile->cf_next)
	{
	if (cfile->cf_user == user  &&  (cfile->cf_flags & FF_ATJOB) != 0)
	    filecount++;		/* Count at jobs for user */
	}
    if (user->cu_entlim != 0  &&  filecount >= user->cu_entlim
      &&  (flags & FF_ATJOB) != 0)
	{
	/*
	 * Tell the user he had too many at jobs
	 */
	DBPRINT1 (8, "addfile:  too many at jobs for user %s\n",
	  user->cu_name + EL_LOGNAME);
	(void) strcpy (errname, TEMPBASE);
	errname[TEMPTYPELOC] = 'f';
	(void) tempfile (errname);
	errfile = fopen (errname, "w");
	if (errfile == NULL)
	    {
	    saveerr = errno;
	    (void) chdir (LIBDIR);
	    (void) fseek (stderr, 0L, 2);
	    (void) fprintf (stderr,
	      "addfile:  couldn't open error file %s to complain about entry count: %s\n",
	      errname, errmsg (saveerr));
	    return;
	    }
	(void) fprintf (errfile, "To: %s\nSubject:  Too many at jobs\n\n",
	  user->cu_name + EL_LOGNAME);
	decodeatfile (name, &year, &month, &day, &hour, &minute, NULL);
	(void) fprintf (errfile,
	  "One of your at jobs for %4.4d/%2.2d/%2.2d at %2.2d:%2.2d was not queued\n",
	  year, month, day, hour, minute);
	(void) fprintf (errfile,
	  "because you already had too many at jobs scheduled (your limit is %d).\n",
	  user->cu_entlim);
	(void) fclose (errfile);
	mailto (user, errname);
	(void) unlink (errname);
	return;
	}
    newfile = (struct cronfile *) malloc (sizeof (struct cronfile));
    if (newfile == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "addfile:  out of memory for %s\n", name);
	return NULL;
	}
    newfile->cf_user = user;
    newfile->cf_name = strcasn (name, "");	/* Set up file name */
    if (newfile->cf_name == NULL)
	{
	free (newfile);
	return NULL;
	}
    newfile->cf_rtime = 0;
    newfile->cf_flags = flags;
    filehead.cf_prev->cf_next = newfile;
    newfile->cf_prev = filehead.cf_prev;
    filehead.cf_prev = newfile;
    newfile->cf_next = &filehead;
    return newfile;
    }

struct cronuser * adduser (uid)		/* Add user to user table */
    int				uid;	/* User ID to be added */
    {
    reg2 struct passwd *	pwent;	/* User's password file entry */
    reg1 struct cronuser *	user;	/* Crontab user entry */

    for (user = userhead.cu_next;  user != &userhead;  user = user->cu_next)
	if (user->cu_uid == uid)
	    return user;
    pwent = getpwuid (uid);		/* Find user's /etc/passwd entry */
    if (pwent == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "adduser:  couldn't find password entry for uid %d\n", uid);
	return NULL;
	}
    user = (struct cronuser *) malloc (sizeof (struct cronuser));
    if (user == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "adduser:  out of memory\n");
	return NULL;
	}
    user->cu_uid = pwent->pw_uid;
    user->cu_gid = pwent->pw_gid;
    user->cu_flags = 0;
    if ((user->cu_home = strcasn (ENVHOME, pwent->pw_dir)) == NULL)
	{
	free (user);
	return NULL;
	}
    if ((user->cu_name = strcasn (ENVLOGNAME, pwent->pw_name)) == NULL)
	{
	free (user->cu_home);
	free (user);
	return NULL;
	}
    user->cu_queue.cq_next = &(user->cu_queue);
    user->cu_queue.cq_prev = &(user->cu_queue);
	    /* user->cu_queue.cq_entp is unused */
    user->cu_entsizelim = userdefaults.cu_entsizelim;
    user->cu_mailsizelim = userdefaults.cu_mailsizelim;
    user->cu_entlim = userdefaults.cu_entlim;
    user->cu_procs = 0;
    user->cu_proclim = userdefaults.cu_proclim;
    user->cu_qcount = 0;
    user->cu_qlim = userdefaults.cu_qlim;
    user->cu_nice = userdefaults.cu_nice;
    /*
     * We have now created the new user structure.  Link it into the chain
     * and return it.
     */
    userhead.cu_prev->cu_next = user;
    user->cu_prev = userhead.cu_prev;
    userhead.cu_prev = user;
    user->cu_next = &userhead;
    return user;
    }

char * strcasn (str1, str2)		/* Copy & cat strs to alloc'ed area */
    reg2 char *			str1;	/* First string to be copied */
    reg3 char *			str2;	/* Second string to be appended */
    {
    reg1 char *			retstr;	/* String to return */

    retstr = malloc ((unsigned) (strlen (str1) + strlen (str2) + 1));
    if (retstr == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "strcasn:  out of memory\n");
	return NULL;
	}
    return strcat (strcpy (retstr, str1), str2);
    }

/*
 * This routine deletes the entries associated with a given file by marking
 * them as deleted (EF_DELETED).  Note that it does NOT delete the cronfile
 * structure itself, and thus is somewhat misnamed.
 */
void delfile (filep)			/* Delete entries from a given file */
    reg2 struct cronfile *	filep;	/* File description structure */
    {
    reg1 struct cronent *	entp;	/* Current command entry */

    DBPRINT1 (1, "delfile (%s)\n", filep->cf_name);
    for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	{
	if (entp->ce_file == filep)
	    {
#ifdef DEBUG
	    if (debug >= 11)
		{
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr, "Deleting crontab entry:\n");
		dumpent (entp);
		}
#endif
	    entp->ce_flags |= EF_DELETED;
	    }
	}
    }

void readall (reftime)			/* Read all crontab files */
    time_t			reftime; /* Time to reference them to */
    {
    reg1 struct cronfile *	filep;	/* File description structure */
    long			now;	/* Current time */

    DBPRINT0 (1, "readall ()\n");
    /* Get rid of the old file entries, so that they can be reclaimed. */
    for (filep = filehead.cf_next;
      filep != &filehead;
      filep = filep->cf_next)
	delfile (filep);
    /* Reclaim what we just marked deleted. */
    collect ();
    /*
     * Recalculate the year base.  We can do this now because all entries
     * are marked deleted.  By recalculating the year base, we extend the
     * range of the tables when cron runs across a year change.
     */
    now = time ((long *) NULL);
    yearbase = (localtime (&now))->tm_year + 1900;
    /*
     * Now actually read the new files in
     */
    for (filep = filehead.cf_next;
      filep != &filehead;
      filep = filep->cf_next)
	readfile (filep, 1, 0, reftime);
    }

/* 
 * Read in a crontab file.  A lot of checking is done to make sure the
 * file is valid.  The current directory is left in LIBDIR.
 */
void readfile (filep, force, gcoll, reftime)
    reg3 struct cronfile *	filep;	/* File description structure */
    int				force;	/* NZ to force reading */
    int				gcoll;	/* NZ to garbage collect */
    time_t			reftime; /* Time to reference entries to */
    {
    int				day;	/* Day an at job would have run */
    int				entcount; /* Number of entries user has */
    reg1 struct cronent *	entp;	/* Entry to be added to cronlist */
    reg2 FILE *			fp;	/* Stdio's file descriptor */
    int				hour;	/* Hour an at job would have run */
#ifdef LOC_CRON
    char *			loc;    /* location for job (only for ATLOC) */
#endif /* LOC_CRON */
    int				minute;	/* Minute an at job would have run */
    int				month;	/* Month an at job would have run */
    int				saveerr; /* Place to save errno b4 clobber */
    struct stat			statbuf; /* Buffer for file status */
    struct cronuser *		userp;	/* Pointer to current user entry */
    int				year;	/* Year an at job would have run */

    DBPRINT2 (4, "readfile (%s, %d)\n", filep->cf_name, force);

    userp = filep->cf_user;
    if (filep->cf_flags & FF_DELETED)
	{
	DBPRINT0 (8, "readfile:  file has been marked deleted\n");
	return;
	}
    else if (filep->cf_flags & FF_ATJOB)
	{
	DBPRINT0 (8, "readfile:  file is an at job\n");
	/*
	 * The following entry-limit check may seem like just paranoia,
	 * because addfile already did the same thing for us.  But
	 * addfile didn't enforce the limit precisely (because it
	 * didn't know how many cron jobs were active), and when
	 * dealing with security issues it's probably better to go
	 * overboard.
	 */
	for (entcount = 0, entp = enthead.ce_next;
	  entp != &enthead;
	  entp = entp->ce_next)
	    {
	    if (entp->ce_file->cf_user == userp
	      &&  (entp->ce_flags & (EF_MARKDEL | EF_DELETED)) == 0)
		entcount++;		/* Count current entries for user */
	    }
	if (userp->cu_entlim != 0  &&  entcount >= userp->cu_entlim)
	    {
	    /*
	     * Tell the user he had too many entries
	     */
	    DBPRINT0 (8, "readfile:  too many at/cron jobs\n");
	    (void) strcpy (errname, TEMPBASE);
	    errname[TEMPTYPELOC] = 'f';
	    (void) tempfile (errname);
	    errfull = 0;
	    errfile = fopen (errname, "w");
	    if (errfile == NULL)
		{
		saveerr = errno;
		(void) chdir (LIBDIR);
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
		  "readfile:  couldn't open error file %s to complain about entry count: %s\n",
		  errname, errmsg (saveerr));
		return;
		}
	    (void) fprintf (errfile, "To: %s\nSubject:  Too many entries\n\n",
	      userp->cu_name + EL_LOGNAME);
	    decodeatfile (filep->cf_name, &year, &month, &day, &hour, &minute,
	       NULL);
	    (void) fprintf (errfile,
	      "One of your at jobs for %4.4d/%2.2d/%2.2d at %2.2d:%2.2d was not queued\n",
	      year, month, day, hour, minute);
	    (void) fprintf (errfile,
	      "because you already had too many at or cron jobs scheduled (your limit is %d).\n",
	      userp->cu_entlim);
	    (void) fclose (errfile);
	    mailto (userp, errname);
	    (void) unlink (errname);
	    filep->cf_flags |= (FF_MARKDEL | FF_DELETED);
	    collect ();			/* Force coll. to preserve space */
	    return;
	    }
	/*
	 * If this is an at-job file, we just build the sole entry right now.
	 */
	filep->cf_flags &= ~FF_MARKDEL;
	filep->cf_rtime = time ((long *) NULL);
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "!  %s %s/%s queued %s\n",
	  userp->cu_name + EL_LOGNAME, ATJOBDIR, filep->cf_name,
	  timenonl (filep->cf_rtime));
	makeatent (filep, reftime);
	return;
	}
    (void) chdir (CRONTABDIR);
    if (stat (filep->cf_name, &statbuf) < 0)
	{				/* Oops!  It went away */
	saveerr = errno;
	(void) chdir (LIBDIR);
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "readfile:  couldn't stat %s:  %s\n",
	  filep->cf_name, errmsg (saveerr));
	delfile (filep);
	if (gcoll)
	    collect ();
	return;
	}
    /*
     * If we are not forcing the read, and if the file has been read since
     * the last modification, we don't need to read it again.
     */
    if (!force  &&  filep->cf_rtime > statbuf.st_mtime)
	{
	DBPRINT0 (15, "file is up to date - returning\n");
	(void) chdir (LIBDIR);
	return;				/* We're all up to date here */
	}
    /*
     * The file is out of date.  Read in new entries.
     * But first zap the old ones
     */
    delfile (filep);			/* Get rid of any old entries */
    if (gcoll)
	collect ();			/* Free up what memory we can */
    (void) strcpy (errname, TEMPBASE);	/* Create an error-message file */
    errname[TEMPTYPELOC] = 'f';
    (void) tempfile (errname);
    errfull = 0;
    errfile = fopen (errname, "w");
    if (errfile == NULL)
	{
	saveerr = errno;
	(void) chdir (LIBDIR);
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "readfile:  couldn't open error file %s to read %s:  %s\n",
	  errname, filep->cf_name, errmsg (saveerr));
	return;
	}
    (void) fprintf (errfile, "To: %s\nSubject:  Errors in your cron table\n\n",
      userp->cu_name + EL_LOGNAME);
    (void) fprintf (errfile,
      "Your cron file '%s' had the following errors in it:\n\n",
      filep->cf_name);
    fp = fopen (filep->cf_name, "r");
    curfile = filep->cf_name;
    curline = 1;
    if (fp == NULL)
	{
	saveerr = errno;
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (errfile, "readfile:  couldn't open %s:  \n", curfile,
	  errmsg (saveerr));
	(void) fclose (errfile);
	mailto (userp, errname);
	(void) unlink (errname);
	(void) chdir (LIBDIR);
	return;
	}
    /*
     * Mode must have SUID and IREAD, but not SGID, SVTX, any EXEC, or
     * world write permissions.
     */
    if ((statbuf.st_mode & MUSTHAVE) != MUSTHAVE
      ||  (statbuf.st_mode & CANTHAVE) != 0)
	{
	(void) fprintf (errfile, "readfile:  %s has illegal mode 0%o\n",
	  filep->cf_name, statbuf.st_mode);
	(void) fclose (errfile);
	mailto (userp, errname);
	(void) unlink (errname);
	(void) chdir (LIBDIR);
	return;
	}
    if (statbuf.st_uid != userp->cu_uid)
	{
	(void) fprintf (errfile, "readfile:  %s is not owned by UID %s\n",
	  filep->cf_name, userp->cu_name + EL_LOGNAME);
	(void) fclose (errfile);
	mailto (userp, errname);
	(void) unlink (errname);
	(void) chdir (LIBDIR);
	return;
	}
    if (statbuf.st_gid != userp->cu_gid)
	{
	(void) fprintf (errfile, "readfile:  %s is not owned by GID %s\n",
	  filep->cf_name, ascgid (userp->cu_gid));
	(void) fclose (errfile);
	mailto (userp, errname);
	(void) unlink (errname);
	(void) chdir (LIBDIR);
	return;
	}
    /*
     * We have successfully opened the file, and are prepared to read it.
     */
    filep->cf_rtime = time ((long *) NULL);
    (void) fseek (stderr, 0L, 2);
    (void) fprintf (stderr, "!  %s %s/%s read %s\n",
      userp->cu_name + EL_LOGNAME, CRONTABDIR, filep->cf_name,
      timenonl (filep->cf_rtime));
#ifdef DEBUG
    if (debug >= 6  &&  debug < 9)
	dumpfile (filep);
    if (debug >= 7  &&  debug < 9)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "file\tuid\tgid\tflags\ttime\n");
	}
#endif
    for (entcount = 0, entp = enthead.ce_next;
      entp != &enthead;
      entp = entp->ce_next)
	{
	if (entp->ce_file->cf_user == userp
	  &&  (entp->ce_flags & (EF_MARKDEL | EF_DELETED)) == 0)
	    entcount++;			/* Count current entries for user */
	}
    while ((entp =
      readent (fp, userp)) != NULL)
	{
	entcount++;
	entp->ce_file = filep;
	entp->ce_rtime = nextrun (reftime, entp);
	if (entp->ce_rtime == 0)	/* Too late for this entry? */
	    {
	    DBPRINT0 (7, "too late, discarding entry\n");
	    if (entp->ce_command)
		free (entp->ce_command);
	    free (entp);
	    entcount--;
	    }
	else if (userp->cu_entlim != 0  &&  entcount > userp->cu_entlim)
	    {
	    if (entcount == userp->cu_entlim + 1)
		{
		/*
		 * First extra entry, report it
		 */
		DBPRINT0 (8,
		  "readfile: too many cron entries, discarding some\n");
		(void) fprintf (errfile,
		  "readfile:  you have too many at/cron entries.  All entries after line %d\n",
		  curline);
		(void) fprintf (errfile,
		  "in %s have been discarded (your limit is %d).\n",
		  curfile, userp->cu_entlim);
		errfull = 1;
		}
	    if (entp->ce_command)
		free (entp->ce_command);
	    free (entp);
	    }
	else
	    {
	    insent (entp);
#ifdef DEBUG
	    if (debug >= 7  &&  debug < 9)
		dumpent (entp);
#endif
	    }
	}
    (void) fclose (fp);
    (void) fclose (errfile);
    if (errfull)
	mailto (userp, errname);
    (void) unlink (errname);
#ifdef DEBUG
    if (debug >= 9)
	dumpall ();
#endif
    (void) chdir (LIBDIR);
    }

static struct cronent * readent (file, user) /* Read one entry */
    reg4 FILE *			file;	/* File to read from */
    struct cronuser *		user;	/* User we're reading on behalf of */
    {
    reg5 int			ch;	/* Next character read */
    reg3 char *			buf;	/* Buffer for longer strings */
    reg2 char *			bufp;	/* Pointer into the above */
    reg7 struct group *		grent;	/* Group file entry for gid */
    int				mdaynonwild = 0; /* Old-fmt mday wasn't wild */
    reg1 struct cronent *	newent;	/* Entry we are creating */
    reg8 int			oldformat; /* NZ if old-format entry */
    reg6 struct passwd *	pwent;	/* Password file entry for uid */
    int				wdaynonwild = 0; /* Old-fmt wday wasn't wild */

    newent = (struct cronent *) malloc (sizeof (struct cronent));
    if (newent == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr, "readent:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile, "readent:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	return NULL;
	}
#ifdef LOC_CRON
    newent->ce_location = NULL;		/* For safety */
#endif /* LOC_CRON */
    newent->ce_uid = user->cu_uid;
    newent->ce_gid = user->cu_gid;
    newent->ce_flags = 0;
    for (  ;  ;  )
	{
	ch = getc (file);
	if (ch == EOF)			/* Return NULL on EOF */
	    {
	    free (newent);
	    return NULL;
	    }
	else if (ch == '\n')
	    curline++;
	else if (ch == '#')		/* Comment? */
	    {
	    while (getc (file) != '\n')
		;			/* Yes, skip it */
	    curline++;
	    }
	else if (!isspace (ch))		/* No, whitespace? */
	    break;			/* If not, we've got real data! */
	}
    /*
     * We now have found the beginning of a line.  Start collecting data.
     */
    (void) ungetc (ch, file);
    buf = getnonblank (file);
    if (buf == NULL)
	{
	free (newent);
	return NULL;
	}
    /*
     * Old-format cron table entries start with an asterisk, or more
     * commonly with a digit.  Otherwise, the entry starts with a flag
     * field.
     */
    oldformat = (buf[0] == '*'  ||  isdigit (buf[0]));
    if (oldformat)
	DBPRINT0 (12, "old-format cron entry\n");
    else
	{
	for (bufp = buf;  *bufp;  bufp++)	/* Parse flags */
	    {
	    switch (*bufp)
		{
		case '-':		/* No flags */
		    if (bufp != buf  ||  buf[1] != '\0')
			synerr ("dash ('-') must stand alone in flags");
		    buf[1] = '\0';	/* Ensure loop stops */
		    break;
		case 'w':
		    if (user->cu_uid  &&  user->cu_uid != -1)
			{
			synerr (
			  "w is not allowed in user crontabs, s assumed");
			newent->ce_flags |= EF_SEQSPAWN;
			}
		    else
			newent->ce_flags |= EF_WAIT;
		    break;
		case '&':
		    newent->ce_flags |= EF_SPAWN;
		    break;
		case 's':
		    newent->ce_flags |= EF_SEQSPAWN;
		    break;
		case 'c':
		    newent->ce_flags |= EF_ONLYCATCHUP;
		    break;
		case 'a':
		    newent->ce_flags |= EF_RUNALL;
		    break;
#ifdef LOC_CRON
#ifdef LOC_MOVE
		case 'm':
		    newent->ce_flags |= EF_ONLY_ON_MOVE;
		    break;
#endif /* LOC_MOVE */
#endif /* LOC_CRON */
		case 'u':
		    if (user->cu_uid)
			synerr ("u flag illegal in user crontabs");
		    else
			user->cu_uid = -1;
		    break;
		case 'g':
		    if (user->cu_uid  &&  user->cu_uid != -1)
			synerr ("g flag illegal in user crontabs");
		    else
			user->cu_gid = -1;
		    break;
		default:
		    (void) fprintf (errfile,
		      "%s line %d:  unrecognized flag '%s'\n",
		      curfile, curline, printable (*bufp));
		    errfull = 1;
		    break;
		}
	    }
	free (buf);
	buf = parseprep (file, NULL);
	if (buf == NULL)
	    {
	    free (newent);
	    return NULL;
	    }
	}
    /*
     * Check to see if any illegal flag combinations have been specified
     */
    ch = newent->ce_flags;		/* ch is just a handy temp var */
    ch &= (EF_WAIT | EF_SPAWN | EF_SEQSPAWN);
    if (ch != 0  &&  ch != EF_WAIT  &&  ch != EF_SPAWN  &&  ch != EF_SEQSPAWN)
	{
	synerr ("only one of &, w, and s may appear");
	if (ch & EF_SEQSPAWN)
	    newent->ce_flags &= ~(EF_WAIT | EF_SPAWN);
	else
	    newent->ce_flags &= ~EF_WAIT;
	}
    if ((newent->ce_flags & EF_RUNALL)  &&  ch == 0)
	{
	newent->ce_flags |= EF_WAIT;
	ch |= EF_WAIT;
	}
    if ((newent->ce_flags & EF_ONLYCATCHUP)  &&  ch == 0)
	newent->ce_flags |= EF_SEQSPAWN;
    if (user->cu_uid  &&  user->cu_uid != -1)
	{
	if ((newent->ce_flags & (EF_RUNALL | EF_SPAWN))
	  == (EF_RUNALL | EF_SPAWN))
	    {
	    synerr ("& and a are not allowed together in user crontabs");
	    newent->ce_flags &= ~EF_RUNALL;
	    }
	}
    /*
     * We are now ready to parse the time fields.
     *
     * The old format is:
     *
     * minute hour monthday month weekday
     *
     * Old-format fields can be numeric or symbolic, and accept a
     * slash-preceded increment.
     *
     * The new format is:
     *
     * [weekday] [month] [monthday] [year] [hour]:[minute]
     *
     * "Weekday" and "month" are specified symbolically, using a
     * case-insensitive three-letter code from "daynames" or "monames".
     * All other fields are numeric, except that the hour can be suffixed
     * with "A", "P", "AM", or "PM", to indicate 12-hour instead of
     * 24-hour time.  The year is identified by virtue of being four
     * digits;  the time is identified by the required colon.  Note that
     * the minimum time specification is a single colon;  this specifies
     * every minute and probably isn't very useful.
     *
     * Any field, whether numeric or not, can be specified as a range of
     * values or a comma-separated list of ranges.  A range is a pair
     * of dash-separated ("-") values, optionally followed by a slash ("/")
     * and an increment.  If the first value in a range is greater than the
     * last, it "wraps around" the maximum value (this is not available for
     * years).
     *
     * For example, the specification "Mar-Aug/2,Oct,Dec-Jan" means the
     * same as "jan,mar,may,jul,oct,dec".
     *
     * Any omitted field is taken to mean a wildcard (all values).  In
     * addition, for a more pleasing format, an asterisk can be used to
     * mean all values for the hour or minute.
     *
     * Finally, any field may be preceded by an exclamation point to indicate
     * that the times NOT selected by the field are to be given.  The
     * exclamation point must be the first character in the field.
     *
     * Examples:
     *
     * :				Every minute
     * Sat-Fri Jan-Dec 1-31 0000-9999 0-23:0-59
     *					The same, the hard way
     * Fri 13				Every minute on Friday the 13th
     * Fri 13:				Every Friday at 13:00 thru 13:59
     * Fri 13:*				The same as above, a bit more readably
     * Fri 13 8P:00			Friday the 13th at 8 PM
     * Tue Nov 1-7 1788-9999/4 8:0	USA presidential election day at 8 AM
     * 8:00				Every day at 8 AM
     * *:15				15 minutes after every hour
     * !mon,Wed-Fri 0:00		Midnight Sunday, Tuesday, and Saturday
     */

    /*
     * First, assume a wildcard value for all fields.
     */
    (void) memset ((char *) newent->ce_wkday, ~0,
      BITSIZE (7) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_month, ~0,
      BITSIZE (12) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_moday, ~0,
      BITSIZE (31) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_year, ~0,
      BITSIZE (YEARMAX) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_hour, ~0,
      BITSIZE (24) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_min, ~0,
      BITSIZE (60) * sizeof (MASKTYPE));
#ifdef LOC_CRON
#ifdef LOC_MOVE
    newent->ce_flags &= ~(EF_IS_AT_LOCATION | EF_WAS_AT_LOCATION);
#else /* ! LOC_MOVE */
    newent->ce_flags &= ~EF_IS_AT_LOCATION;
#endif /* LOC_MOVE */
#ifdef LOC_EVERY
    newent->ce_every_spec = NULL;
    newent->ce_every_next = 0;
#endif /* LOC_EVERY */
#endif
    /*
     * Get the first field
     */
    if (oldformat)
	{
	/*
	 * Minute
	 */
	getnfield (buf, newent->ce_min, 60, 0, 0);
	buf = parseprep (file, buf);
	}
    else
	{
	/*
	 * Now, see if we have a weekday specification
	 */
	bufp = buf;
	if (*bufp == '!')
	    bufp++;
	if (prefixnum (bufp, daynames, 7) >= 0)
	    {				/* Weekday was specified */
	    getafield (buf, newent->ce_wkday, daynames, 7);
	    buf = parseprep (file, buf);
	    }
	}
    if (buf == NULL)
	{
	free (newent);
	return NULL;
	}
    if (oldformat)
	{
	/*
	 * Hour
	 */
	getnfield (buf, newent->ce_hour, 24, 0, 0);
	buf = parseprep (file, buf);
	}
    else
	{
	/*
	 * See if we have a month specification
	 */
	bufp = buf;
	if (*bufp == '!')
	    bufp++;
	if (prefixnum (bufp, monames, 12) >= 0)
	    {				/* Month was specified */
	    getafield (buf, newent->ce_month, monames, 12);
	    buf = parseprep (file, buf);
	    }
	}
    if (buf == NULL)
	{
	free (newent);
	return NULL;
	}
    /*
     * See if we have a day-of-month specification.  In old-format
     * tables, this is required.
     */
    if (oldformat)
	{
	mdaynonwild = buf[0] != '*';
	getnfield (buf, newent->ce_moday, 31, 1, 0);
	buf = parseprep (file, buf);
	}
    else
	{
	bufp = buf;
	if (*bufp == '!')
	    bufp++;
	for (  ;  isdigit (*bufp);  bufp++)	/* Find non-digit */
	    ;
	if ((bufp - buf == 1  ||  bufp - buf == 2) /* 1 or two digits? */
	  &&  strchr (buf, ':') == NULL)	/* ..but no colon in field? */
	    {					/* ..then day was specified */
	    getnfield (buf, newent->ce_moday, 31, 1, 0);
	    buf = parseprep (file, buf);
	    }
	}
    if (buf == NULL)
	{
	free (newent);
	return NULL;
	}
    bufp = buf;
    if (*bufp == '!')
	bufp++;
    if (oldformat)
	{
	/*
	 * Month.  This can be symbolic or numeric.
	 */
	if (prefixnum (bufp, monames, 12) >= 0)
	    getafield (buf, newent->ce_month, monames, 12); /* Symbolic */
	else
	    getnfield (buf, newent->ce_month, 12, 0, 0);
	buf = parseprep (file, buf);
	}
    else
	{
	/*
	 * See if we have a year specification
	 */
	for (  ;  isdigit (*bufp);  bufp++)	/* Find non-digit */
	    ;
	if (isspace (*bufp)  &&  bufp - buf == 4) /* Exactly four digits? */
	    {					/* ..then year was specified */
	    getnfield (buf, newent->ce_year, YEARMAX, yearbase, 1);
	    buf = parseprep (file, buf);
	    }
	}
    if (buf == NULL)
	{
	free (newent);
	return NULL;
	}
    if (oldformat)
	{
	/*
	 * Weekday.  Again, this can be symbolic or numeric.
	 */
	wdaynonwild = buf[0] != '*';
	bufp = buf;
	if (*bufp == '!')
	    bufp++;
	if (prefixnum (bufp, daynames, 7) >= 0)
	    getafield (buf, newent->ce_wkday, daynames, 7);
	else
	    getnfield (buf, newent->ce_wkday, 7, 0, 0);
	}
    else
	{
	/*
	 * At this point, we had better have a hour/minute specification.
	 */
	bufp = strchr (buf, ':');		/* Look for colon */
#ifdef LOC_CRON
#ifdef LOC_MOVE
	/* Edge detection doesn't require any time spec */
	if ((newent->ce_flags & EF_ONLY_ON_MOVE)  &&  *buf == '@')
	    {
	    /* save location info in this special case */
	    newent->ce_location = strcasn("", buf + 1);
	    free(buf);
	    }
	else
#endif /* LOC_MOVE */
#endif /* LOC_CRON */
            {
	    if (bufp == NULL)			/* Colon better be there */
	        {
	        synerr ("couldn't find hour:minute field");
		free (newent);
		return NULL;
	        }
	    *bufp++ = '\0';
	    getnfield (buf, newent->ce_hour, 24, 0, 0);
	    getnfield (bufp, newent->ce_min, 60, 0, 0);
	    free (buf);
	    }
	}
#ifdef LOC_CRON
    /*
     * Location information.
     * Basic format: @place
     * Place is actually a regular expression.
     */
    /* NEEDSWORK: doesn't necessarily handle oldformat case */
    while ((ch = getc(file)) != EOF)
	{
	if (ch == '@' || !isspace(ch))
	    break;
	}
    if(ch == '@')
	{
	ch = getc(file);
	if(!isspace(ch))
	    {
	    char *my_buf = NULL;
	    (void)ungetc(ch, file);
	    my_buf = parseprep(file, my_buf);
	    if (my_buf == NULL)
		{
		if (newent->ce_location)
		    free (newent->ce_location);
		free (newent);
		return NULL;
		}
	    newent->ce_location = my_buf;
	    DBPRINT1(12, "location for new entry %s\n", newent->ce_location);
	    }
	}
    else
	(void) ungetc(ch,file);
    skipblanks (file);

#ifdef LOC_MOVE
    if ((newent->ce_flags & EF_ONLY_ON_MOVE) && newent->ce_location == NULL)
	{
	synerr ("m option must have location specification");
	if (newent->ce_location)
	    free (newent->ce_location);
	free (newent);
	return NULL;
	}
#endif /* LOC_MOVE */

#ifdef LOC_EVERY
    while ((ch = getc(file))!=EOF)
	{
	if (ch == '>' || !isspace(ch))
	    break;
	}
    if (ch == '>')
	{
	ch = getc(file);
	if (!isspace(ch))
	    {
	    (void)ungetc(ch, file);
	    my_buf = parseprep(file, my_buf);
	    if (my_buf == NULL)
		{
		if (newent->ce_location)
		    free (newent->ce_location);
		free (newent);
		return NULL;
		}
	    newent->ce_every_spec = my_buf;
	    newent->ce_every_next = parse_every(my_buf);
	    DBPRINT1(12, "every for new entry %s\n", newent->ce_every_spec);
	    }
	}
    else
	(void) ungetc(ch,file);
    skipblanks (file);
#endif /* LOC_EVERY */
#else
    skipblanks (file);
#endif
    if (user->cu_uid == -1)		/* u flag specified? */
	{
	buf = getnonblank (file);
	if (buf == NULL)
	    {
#ifdef LOC_CRON
	    if (newent->ce_location)
		free (newent->ce_location);
#endif /* LOC_CRON */
	    free (newent);
	    return NULL;
	    }
	skipblanks (file);
	if (isdigit (buf[0]))
	    newent->ce_uid = atoi (buf);
	else
	    {
	    pwent = getpwnam (buf);
	    if (pwent == NULL)
		{
		(void) fprintf (errfile,
		  "%s line %d:  uid '%s' is unrecognized\n",
		  curfile, curline, buf);
		errfull = 1;
#ifdef LOC_CRON
		if (newent->ce_location)
		    free (newent->ce_location);
#endif /* LOC_CRON */
		free (newent);
		return NULL;
		}
	    else
		newent->ce_uid = pwent->pw_uid;
	    }
	}
    if (user->cu_gid == -1)		/* g flag specified? */
	{
	buf = getnonblank (file);
	if (buf == NULL)
	    {
#ifdef LOC_CRON
	    if (newent->ce_location)
		free (newent->ce_location);
#endif /* LOC_CRON */
	    free (newent);
	    return NULL;
	    }
	skipblanks (file);
	if (isdigit (buf[0]))
	    newent->ce_gid = atoi (buf);
	else
	    {
	    grent = getgrnam (buf);
	    if (grent == NULL)
		{
		(void) fprintf (errfile,
		  "%s line %d:  gid '%s' is unrecognized\n",
		  curfile, curline, buf);
		errfull = 1;
#ifdef LOC_CRON
		if (newent->ce_location)
		    free (newent->ce_location);
#endif /* LOC_CRON */
		free (newent);
		return NULL;
		}
	    else
		newent->ce_gid = grent->gr_gid;
	    }
	}
    buf = getcommand (file, user->cu_entsizelim);
    if (buf == NULL)
	{
#ifdef LOC_CRON
	if (newent->ce_location)
	    free (newent->ce_location);
#endif /* LOC_CRON */
	free (newent);
	return NULL;
	}
    newent->ce_command = buf;
    newent->ce_input = strchr (newent->ce_command, '\n');
    if (newent->ce_input != NULL)
	*(newent->ce_input++) = '\0';
    if (oldformat  &&  mdaynonwild  &&  wdaynonwild)
	synerr ("old-format cron line needs rewriting; see COMPATIBILITY in 'man crontab'");
    return newent;
    }

static void makeatent (file, reftime)	/* Make an at-job entry */
    reg1 struct cronfile *	file;	/* File containing at commands */
    time_t			reftime; /* Time to reference the entry to */
    {
    int				day;	/* Day to run */
    int				hour;	/* Hour to run */
    char *			loc = NULL; /* location for job (only for */
					/* ..ATLOC) */
    int				minute;	/* Minute to run */
    int				month;	/* Month to run */
    reg2 struct cronent *	newent;	/* Entry we are creating */
    int				year;	/* Year to run */

    DBPRINT1 (1, "makeatent (%s)\n", file->cf_name);
    newent = (struct cronent *) malloc (sizeof (struct cronent));
    if (newent == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "makeatent:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile,
	  "makeatent:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	return;
	}
    file->cf_rtime = time ((time_t *) NULL);
    newent->ce_uid = file->cf_user->cu_uid;
    newent->ce_gid = file->cf_user->cu_gid;
    newent->ce_flags = 0;
    newent->ce_file = file;
    newent->ce_command = NULL;
#ifdef LOC_CRON
#ifdef LOC_MOVE
    newent->ce_flags &=
      ~(EF_IS_AT_LOCATION | EF_WAS_AT_LOCATION | EF_ONLY_ON_MOVE | EF_ONLY_ON_MOVE_FIRED);
#else /* ! LOC_MOVE */
    newent->ce_flags &= ~EF_IS_AT_LOCATION;
#endif /* LOC_MOVE */
    newent->ce_location = NULL;
#ifdef LOC_EVERY
    newent->ce_every_spec = NULL;
    newent->ce_every_next = 0;
#endif /* LOC_EVERY */
#endif
    newent->ce_input = NULL;
    newent->ce_flags = EF_ATJOB;
    /*
     * Set up the time specification.  The weekday is always wildcarded;
     * the other fields are set to zero and will have single bit set later.
     */
    (void) memset ((char *) newent->ce_wkday, ~0,
      BITSIZE (7) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_month, 0,
      BITSIZE (12) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_moday, 0,
      BITSIZE (31) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_year, 0,
      BITSIZE (YEARMAX) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_hour, 0,
      BITSIZE (24) * sizeof (MASKTYPE));
    (void) memset ((char *) newent->ce_min, 0,
      BITSIZE (60) * sizeof (MASKTYPE));
    decodeatfile (file->cf_name, &year, &month, &day, &hour, &minute, &loc);
    month--;				/* Zero-index the one-indexed stuff */
    day--;				/* ... */
    if (year < yearbase)
	{
	/*
	 * Long-past job.  Make it runnable right away.
	 */
	newent->ce_rtime = reftime;
	}
    else if (year >= yearbase + YEARMAX)
	{
	/*
	 * Job far in the future.  Make it disappear.
	 */
	newent->ce_flags |= EF_DELETED;	/* Get rid of obsolete entry */
	re_collect = 1;			/* Force a garbage collection */
	if (loc != NULL)
	    free (loc);
	}
    else
	{
	/*
	 * Job is in a reasonable year range.  Find its runnable time.
	 */
	year -= yearbase;			/* ... */
	newent->ce_year[MASKBYTE (year)] |= MASKBIT (year);
	newent->ce_month[MASKBYTE (month)] |= MASKBIT (month);
	newent->ce_moday[MASKBYTE (day)] |= MASKBIT (day);
	newent->ce_hour[MASKBYTE (hour)] |= MASKBIT (hour);
	newent->ce_min[MASKBYTE (minute)] |= MASKBIT (minute);
#ifdef LOC_CRON
#ifdef LOC_ATLOC
	if (loc != NULL)
	    {
	    if (year == ATLOC_NOYEAR)
		newent->ce_flags |= EF_ONLY_ON_MOVE;
	    newent->ce_location = loc;  /* (already strdup'ed) */
	    }
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
	newent->ce_rtime = nextrun (reftime, newent);
	}
    if (newent->ce_rtime == 0)		/* Too far in future? */
	{
	newent->ce_flags |= EF_DELETED;	/* Get rid of obsolete entry */
	re_collect = 1;			/* Force a garbage collection */
	if (loc != NULL)
	    free (loc);
	}
    /*
     * Link the new entry into the entry list
     */
    insent (newent);
#ifdef DEBUG
    if (debug >= 10)
	dumpent (newent);
#endif
    }

/*
 * extract the at stuff from the file for type-b files
 */
#ifdef LOC_CRON
#ifdef LOC_ATLOC
static void atloc_decode_at_file (name, year, month, day, hour, minute, locp)
    reg2 char *			name;	/* Name of the file holding the job */
    reg1 int *			year;	/* Year job should run */
    int *			month;	/* Month job should run */
    int *			day;	/* Day job should run */
    int *			hour;	/* Hour job should run */
    int *			minute;	/* Minute job should run */
    char **			locp; /* location, only for ATLOC */
    {
    FILE *f;
/* xxx */
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif /* ! MAXPATHLEN */
    char buf[MAXPATHLEN];
    char *bufend;
    char kind;
    int n;

    /*
     * There's a possible buffer overflow here,
     * but since both ATJOBDIR and name
     * are sysadmin or xcron (not user) generated
     * we trust them.
     */
    sprintf(buf, "%s/%s", ATJOBDIR, name);
    if ((f = fopen(buf, "r")) == NULL) {
	    fprintf(stderr, "atloc_decode_at_file: internal problem, cannot read file %s.\n", name);
	    return;
    };
    /* Must have defaults for everything since something
     * may be missing.  Pick a date in the past so we can
     * handle edge-triggered jobs.
     */
    *year = ATLOC_NOYEAR;
    *month = 1;
    *day = 1;
    *hour = 0;
    *minute = 0;
    if (locp)
	    *locp = NULL;
    while (fgets(buf, sizeof buf, f)) {
	    if (buf[0] != '#') /* everything hidden in shell comments */
		    break;
	    if (buf[1] == '.') /* end of attributes */
		    break;
	    kind = buf[1];
	    /* perl's chomp */
	    bufend = &buf[strlen(buf) - 1];
	    if (*bufend == '\n')
		    *bufend = 0;
	    n = atoi(&buf[2]); /* just in case */
	    switch (kind) {
	    case 'Y': /* 4-digit year */
		    *year = n;
		    break;
	    case 'm': /* month (1-based) */
		    *month = n;
		    break;
	    case 'd': /* day of month (1-based) */
		    *day = n;
		    break;
	    case 'H': /* 24 hour */
		    *hour = n;
		    break;
	    case 'M': /* minute */
		    *minute = n;
		    break;
	    case '@': /* location */
		    if (!locp)  /* can pass null locp */
			    break;
		    if (*locp)
			    free(*locp);
		    *locp = strcasn ("", &buf[2]);
		    break;
	    default:
		    fprintf(stderr, "atloc_decode_at_file: internal problem, unknown kind %c on line %s.\n", kind, buf);
		    break;
	    };
    };
    fclose(f);
    }
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */

/* 
 * Turn an at-job name into a broken-out time.  Returns the time in
 * the pointed-to variables.  The month and day are one-indexed, the
 * hour is 24-hour, and the year is 4-digit.
 */
static void decodeatfile (name, year, month, day, hour, minute, locp)
    reg2 char *			name;	/* Name of the file holding the job */
    reg1 int *			year;	/* Year job should run */
    int *			month;	/* Month job should run */
    int *			day;	/* Day job should run */
    int *			hour;	/* Hour job should run */
    int *			minute;	/* Minute job should run */
    char **			locp; /* location, only for ATLOC */
    {
#ifdef LOC_CRON
#ifdef LOC_ATLOC
    /*
     * atloc format starts with 'b'
     * and encodes time an location IN the file itself.
     * (What a concept.)
     */
    if (name[WHEN_FLAG] == ATLOC_FLAG_VALUE) {
	atloc_decode_at_file(name, year, month, day, hour, minute, locp);
	return;
    };
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
#ifdef LONGNAMES
    /*
     * There are two file formats.  The first is
     * YYMMDDhhmmss<garbage>; this was used on older systems with
     * 14-character filenames.  The second is
     * aYYYYMMDDhhmmss<garbage>, where the "a" is literal.  This is
     * used on current systems.
     */
    if (name[WHEN_FLAG] == WHEN_FLAG_VALUE)
	{
	(void) sscanf (&name[WHEN_YEAR], "%4d%2d%2d%2d%2d",
	  year, month, day, hour, minute);
	}
    else
	{
	(void) sscanf (&name[WHEN_OLD_YEAR], "%2d%2d%2d%2d%2d",
	  year, month, day, hour, minute);
	if (*year < Y2K_THRESHOLD)
	    *year += Y2K_2000;		/* Correct for year-2000 problems */
	else
	    *year += Y2K_1900;
	}
#else /* LONGNAMES */
    /* The filename format is YYMMDDhhmmss<garbage> */
    (void) sscanf (&name[WHEN_YEAR], "%2d%2d%2d%2d%2d",
      year, month, day, hour, minute);
    if (*year < Y2K_THRESHOLD)
	*year += Y2K_2000;		/* Correct for year-2000 problems */
    else
	*year += Y2K_1900;
#endif /* LONGNAMES */
    }

void insent (newent)			/* Insert an entry into the list */
    reg2 struct cronent *	newent;	/* Entry to insert */
    {
    reg1 struct cronent *	entp;	/* Current entry */

    if (newent->ce_flags & EF_DELETED)
	entp = &enthead;		/* Insert at tail if deleted */
    else
	{
	for (entp = enthead.ce_next;  entp != &enthead;  entp = entp->ce_next)
	    {
	    if (newent->ce_rtime < entp->ce_rtime)
		break;			/* We belong in front of entp */
	    }
	}
    entp->ce_prev->ce_next = newent;
    newent->ce_next = entp;
    newent->ce_prev = entp->ce_prev;
    entp->ce_prev = newent;
    }

static char * getnonblank (file)	/* Get a nonblank string from a file */
    reg5 FILE *			file;	/* File to read from */
    {
    reg2 int			ch;	/* Character read from file */
    reg3 char *			buf;	/* Buffer to fill up */
    reg1 char *			bufp;	/* Pointer into the buffer */
    reg4 int			bufsize; /* Current size of the buffer */

    bufsize = 20;
    bufp = buf = malloc ((unsigned) bufsize);
    if (buf == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "getnonblank:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile,
	  "getnonblank:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	return NULL;
	}
    while ((ch = getc (file)) != EOF)
	{
	if (ch == '\n')
	    {
	    curline++;
	    break;
	    }
	else if (isspace (ch))
	    break;
	else
	    *bufp++ = ch;
	if (bufp - buf >= bufsize - 1)
	    {
	    buf = realloc (buf, (unsigned) bufsize + 80);
	    if (buf == NULL)
		{
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
		  "getnonblank:  out of memory for %s at line %d\n",
		  curfile, curline);
		(void) fprintf (errfile,
		  "getnonblank:  out of memory for %s at line %d\n",
		  curfile, curline);
		errfull = 1;
		return NULL;
		}
	    bufp = buf + bufsize - 1;
	    bufsize += 80;
	    }
	}
    *bufp++ = '\0';
    DBPRINT1 (12, "getnonblank:  returning '%s'\n", buf);
    buf = realloc (buf, (unsigned) (bufp - buf));
    if (buf == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "getnonblank:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile,
	  "getnonblank:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	}
    return buf;
    }

static char * getcommand (file, limit)	/* Get a command string from a file */
    reg5 FILE *			file;	/* File to read from */
    reg6 int			limit;	/* Limit on size of command string */
    {
    reg2 int			ch;	/* Character read from file */
    reg3 char *			buf;	/* Buffer to fill up */
    reg1 char *			bufp;	/* Pointer into the buffer */
    reg4 int			bufsize; /* Current size of the buffer */
    int				overlimit = 0; /* NZ if limit exceeded */

    bufsize = 20;
    bufp = buf = malloc ((unsigned) bufsize);
    if (buf == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "getcommand:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile,
	  "getcommand:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	return NULL;
	}
    while ((ch = getc (file)) != EOF)
	{
	if (ch == '\n')
	    {
	    curline++;
	    break;
	    }
	else if (ch == '%')			/* % becomes newline */
	    *bufp++ = '\n';			/* What a stupid convention */
	else if (ch == '\\')
	    {					/* Handle backslashes */
	    ch = getc (file);
	    if (ch == EOF)			/* \ at EOF passes through */
		{
		*bufp++ = ch;
		break;
		}
	    switch (ch)
		{
		case 'n':			/* \n becomes newline */
		    *bufp++ = '\n';
		    break;
		case '%':			/* \% becomes % */
		case '\\':			/* \\ becomes \ */
		    *bufp++ = ch;
		    break;
		case '\n':			/* \ newline becomes nothing */
		    curline++;
		    /* Backslashed newline.  Skip whitespace at next line's
		     * start.
		     */
		    while ((ch = getc (file)) == ' '  ||  ch == '\t')
			;
		    if (ch != '\\'  &&  ch != EOF)
			(void) ungetc (ch, file);
		    break;
		default:			/* All else passes through */
		    *bufp++ = '\\';
		    *bufp++ = ch;
		    break;
		}
	    }
	else
	    *bufp++ = ch;
	if (bufp - buf >= limit)
	    {
	    if (!overlimit)
		{
		overlimit = 1;
		(void) fprintf (stderr, "getcommand:  rejecting long line\n");
		(void) fprintf (errfile,
		   "Line %d of %s is too long, ignored.\n",
		   curline, curfile);
		(void) fprintf (errfile,
		  "Your limit is a %d-byte command.  Try using a shell script instead.\n",
		  limit);
		errfull = 1;
		}
	    bufp = buf;
	    }
	if (bufp - buf >= bufsize - 1)
	    {
	    buf = realloc (buf, (unsigned) bufsize + 80);
	    if (buf == NULL)
		{
		(void) fseek (stderr, 0L, 2);
		(void) fprintf (stderr,
		  "getcommand:  out of memory for %s at line %d\n",
		  curfile, curline);
		(void) fprintf (errfile,
		  "getcommand:  out of memory for %s at line %d\n",
		  curfile, curline);
		errfull = 1;
		return NULL;
		}
	    bufp = buf + bufsize - 1;
	    bufsize += 80;
	    }
	}
    if (overlimit)
	{
	free (buf);
	return NULL;
	}
    *bufp++ = '\0';
    DBPRINT1 (12, "getcommand:  returning '%s'\n", buf);
    buf = realloc (buf, (unsigned) (bufp - buf));
    if (buf == NULL)
	{
	(void) fseek (stderr, 0L, 2);
	(void) fprintf (stderr,
	  "getcommand:  out of memory for %s at line %d\n",
	  curfile, curline);
	(void) fprintf (errfile,
	  "getcommand:  out of memory for %s at line %d\n",
	  curfile, curline);
	errfull = 1;
	}
    return buf;
    }

static char * parseprep (file, buf)	/* Prepare to parse new field */
    reg1 FILE *			file;	/* File to be read */
    reg2 char *			buf;	/* Last buffer parsed */
    {

    if (buf)
	free (buf);			/* Get rid of old buffer */
    skipblanks (file);
    buf = getnonblank (file);
    makelower (buf);
    return buf;
    }

static void getafield (buf, field, table, size) /* Parse a symbolic spec */
    reg2 char *			buf;	/* Buffer containing specification */
    reg6 MASKTYPE *		field;	/* Field to modify */
    reg8 char *			table[]; /* Table of legal names */
    reg7 int			size;	/* Number of bits in the field */
    {
    reg1 char *			endp;	/* Pointer to end of a number */
    reg3 int			first;	/* First or only bit of range */
    reg4 int			incr;	/* Increment for filling range */
    reg9 int			invert;	/* NZ to invert the result */
    reg5 int			last;	/* Last bit of range */

    DBPRINT3 (13, "getafield ('%s', field, (%s), %d)", buf, *table, size);
    DBPRINT0 (15, "\n");
    for (first = 0;  first < BITSIZE (size);  first++)
	field[first] = 0;		/* Clear out the field */
    invert = 0;
    if (*buf == '!')			/* Inverted sense? */
	{
	invert = 1;
	buf++;
	}
    while (*buf)
	{
	for (endp = buf;  islower (*endp);  endp++)
	    ;				/* Skip over the first symbol */
	if (*endp == '-')
	    {
	    *endp++ = '\0';
	    first = prefixnum (buf, table, size);
	    DBPRINT2 (15, "range:  %s (%d) to ", buf, first);
	    for (buf = endp;  islower (*endp);  endp++)
		;
	    last = prefixnum (buf, table, size);
	    DBPRINT2 (15, "%s (%d) by ", buf, last);
	    incr = 1;
	    if (*endp == '/')		/* Increment specified? */
		{
		for (buf = ++endp;  isdigit (*endp);  endp++)
		    ;
		incr = atoi (buf);	/* Atoi stops at non-digit character */
		DBPRINT2 (15, "%s (%d)\n", buf, incr);
		}
#ifdef DEBUG
	    else
		DBPRINT0 (15, "1\n");
#endif
	    if (*endp  &&  *endp != ',')
		synerr ("range specification must be followed by comma");
	    if (*endp)
		*endp++ = '\0';		/* Same as tricky section below */
	    if (first < 0)
		{
		(void) fprintf (errfile,
		  "%s line %d:  symbol '%s' unrecognized, '%s' assumed\n",
		  curfile, curline, buf, table[0]);
		errfull = 1;
		first = 0;
		}
	    if (last < 0)
		{
		last = size - 1;
		(void) fprintf (errfile,
		  "%s line %d:  symbol '%s' unrecognized, '%s' assumed\n",
		  curfile, curline, buf, table[last]);
		errfull = 1;
		}
	    if (last < first)	/* Handle wraparound ranges (e.g. 23-2) */
		{
		for (  ;  first < size;  first += incr)
		    field[MASKBYTE (first)] |= MASKBIT (first);
		first %= size;
		}
	    for (  ;  first <= last;  first += incr)
		field[MASKBYTE (first)] |= MASKBIT (first);
	    }
	else
	    {
	    /* The following statement is tricky.  If we are at a zero byte,
	     * we leave endp there, which will break us out of the loop.
	     * Otherwise we create a zero byte for atoi (), and step over it
	     * so subsequent loop passes will see the remaining data.
	     */
	    if (*endp != ','  &&  *endp != '\0')
		{
		(void) fprintf (errfile,
		  "%s line %d:  illegal character '%s'\n",
		  curfile, curline, printable (*endp));
		errfull = 1;
		}
	    if (*endp != '\0')
		*endp++ = '\0';
	    first = prefixnum (buf, table, size);
	    DBPRINT2 (15, " single:  %s (%d)\n", buf, first);
	    if (first < 0)
		{
		(void) fprintf (errfile,
		  "%s line %d:  symbol '%s' unrecognized\n",
		  curfile, curline, buf);
		errfull = 1;
		}
	    else
		field[MASKBYTE (first)] |= MASKBIT (first);
	    }
	buf = endp;
	}
    if (invert)
	for (first = 0;  first < BITSIZE (size);  first++)
	    field[first] ^= ~0;	    /* Invert all bits */
    DBPRINT0 (15, " result");
#ifdef DEBUG
    (void) fseek (stderr, 0L, 2);
    if (debug >= 15)
	(void) fprintf (stderr, " result");
    if (debug >= 14)
	{
	(void) fprintf (stderr, " =");
	for (first = 0;  first < BITSIZE (size);  first++)
	    (void) fprintf (stderr, " %x", field[first] & MASKMASK);
	}
    if (debug >= 13)
	(void) putc ('\n', stderr);
#endif
    }

	/* Parse a numeric time specification */
static void getnfield (buf, field, size, base, errorok)
    reg2 char *			buf;	/* Buffer containing specification */
    reg6 MASKTYPE *		field;	/* Field to modify */
    reg8 int			size;	/* Number of bits in the field */
    reg7 int			base;	/* Base of field's numbering */
    reg10 int			errorok; /* Nonzero to ignore range errors */
    {
    reg1 char *			endp;	/* Pointer to end of a number */
    reg3 int			first;	/* First or only bit of range */
    reg4 int			incr;	/* Increment for a range */
    reg9 int			invert;	/* NZ to invert the result */
    reg5 int			last;	/* Last bit of range */

    DBPRINT4 (13, "getnfield ('%s', field, %d, %d, %s)",
      buf, size, base, errorok ? "ignore errors" : "report errors");
    DBPRINT0 (15, "\n");
    last = (*buf == '*'  ||  *buf == '\0') ? ~0 : 0;
    for (first = 0;  first < BITSIZE (size);  first++)
	field[first] = last;
    if (*buf == '*'  ||  *buf == '\0')
	{
	DBPRINT0 (15, " processing asterisk\n");
	if (buf[0] != 0  &&  buf[1] != '\0')
	    synerr ("asterisk must stand alone\n");
	*buf = '\0';			/* Suppress the loop below */
	}
    invert = 0;
    if (*buf == '!')			/* Inverted sense? */
	{
	invert = 1;
	buf++;
	}
    while (*buf)
	{
	for (endp = buf;  isdigit (*endp)  ||  *endp == '*';  endp++)
	    ;
	if (*endp == '-')
	    {
	    *endp++ = '\0';
	    first = atoi (buf) - base;
	    DBPRINT2 (15, "range:  %s (%d) to ", buf, first + base);
	    for (buf = endp;  isdigit (*endp);  endp++)
		;
	    last = atoi (buf) - base;	/* Atoi stops at non-digit character */
	    DBPRINT2 (15, "%s (%d) by ", buf, last + base);
	    incr = 1;
	    if (*endp == '/')		/* Increment specified? */
		{
		for (buf = ++endp;  isdigit (*endp);  endp++)
		    ;			/* Skip over it */
		incr = atoi (buf);	/* Atoi stops at non-digit character */
		if (incr <= 0)
		    {
		    (void) fprintf (errfile,
		      "%s line %d:  illegal increment %d\n",
		      curfile, curline, incr);
		    errfull = 1;
		    incr = 1;
		    }
		DBPRINT2 (15, "%s (%d)\n", buf, incr);
		}
#ifdef DEBUG
	    else
		DBPRINT0 (15, "1\n");
#endif
	    if (*endp  &&  *endp != ',')
		synerr ("range specification must be followed by comma");
	    if (*endp)
		*endp++ = '\0';		/* Same as tricky section below */
	    if (errorok)
		{
		if (first < 0)
		    {
		    first %= incr;
		    if (first < 0)
			first += incr;
		    }
		if (last >= size)
		    last = size - 1;	/* Last can still be < 0 */
		/* First can still be >= size, and last <= 0 */
		DBPRINT3 (15, "adjusted range is %d to %d by %d\n",
		  first + base, last + base, incr);
		}
	    else if (first < 0  ||  first >= size
	      ||  last < 0  ||  last >= size)
		{
		(void) fprintf (errfile,
		  "%s line %d:  illegal range %d to %d (must be %d to %d)\n",
		  curfile, curline, first + base, last + base,
		  base, size + base - 1);
		errfull = 1;
		if (first < 0)
		    first = 0;
		if (first >= size)
		    first = size - 1;
		if (last < 0)
		    last = 0;
		if (last >= size)
		    last = size - 1;
		}
	    /* Handle wraparound ranges (e.g. 23-2) */
	    if (!errorok  &&  last < first)
		{
		for (  ;  first < size;  first += incr)
		    field[MASKBYTE (first)] |= MASKBIT (first);
		first %= size;
		}
	    for (  ;  first <= last;  first += incr)
		field[MASKBYTE (first)] |= MASKBIT (first);
	    }
	else
	    {
	    /* The following statement is tricky.  If we are at a zero byte,
	     * we leave endp there, which will break us out of the loop.
	     * Otherwise we create a zero byte for atoi (), and step over it
	     * so subsequent loop passes will see the remaining data.
	     */
	    if (*endp != ','  &&  *endp != '\0')
		{
		(void) fprintf (errfile,
		  "%s line %d:  illegal character '%s'\n",
		  curfile, curline, printable (*endp));
		errfull = 1;
		}
	    if (*endp != '\0')
		*endp++ = '\0';
	    first = atoi (buf) - base;
	    DBPRINT2 (15, " single:  %s (%d)\n", buf, first + base);
	    if (first >= 0  &&  first < size)
		field[MASKBYTE (first)] |= MASKBIT (first);
	    else
		{
		if (!errorok)
		    {
		    (void) fprintf (errfile,
		      "%s line %d:  illegal value %d (must be %d to %d)\n",
		      curfile, curline, first + base,
		      base, size + base - 1);
		    errfull = 1;
		    }
		}
	    }
	buf = endp;
	}
    if (invert)
	for (first = 0;  first < BITSIZE (size);  first++)
	    field[first] ^= ~0;	    /* Invert all bits */
#ifdef DEBUG
    (void) fseek (stderr, 0L, 2);
    if (debug >= 15)
	(void) fprintf (stderr, " result");
    if (debug >= 14)
	{
	(void) fprintf (stderr, " =");
	for (first = 0;  first < BITSIZE (size);  first++)
	    (void) fprintf (stderr, " %x", field[first] & MASKMASK);
	}
    if (debug >= 13)
	(void) putc ('\n', stderr);
#endif
    }

static void skipblanks (file)		/* Skip whitespace */
    reg1 FILE *			file;	/* File to skip on */
    {
    reg2 int			ch;	/* Character just read */

    while ((ch = getc (file)) == ' '  ||  ch == '\t')
	;
    (void) ungetc (ch, file);
    }

static void makelower (buf)		/* Make a buffer lowercase */
    reg1 char *			buf;	/* Buffer to modify */
    {
    for (  ;  *buf;  buf++)
	{
	if (isupper (*buf))
	    *buf = tolower (*buf);
	}
    }

static int prefixnum (str, table, tablesize) /* See if str has known prefix */
    char *		str;		/* String to test */
    reg3 char *		table[];	/* Table of prefix pointers */
    reg4 int		tablesize;	/* Size of the table */
    {
    reg2 char *		strch;		/* Pointer to test string */
    reg1 char *		tablech;	/* Pointer to table string */

    while (--tablesize >= 0)
	{
	for (strch = str, tablech = table[tablesize];
	  *tablech  &&  *strch == *tablech;
	  strch++, tablech++)
	    ;
	if (*tablech == '\0')		/* Did we get to end of table ent? */
	    break;			/* If eq yes, return prefix number */
	}
    return tablesize;			/* This is -1 if we failed */
    }
